! Policy simulations for the directed search model
module params
    implicit none
! General
    real*8, parameter :: one=1.0D+0
    real*8, parameter :: zero=0.0D+0
! Grids
    integer, parameter :: npts=250 ! Firm size
    integer, parameter :: nepspts=25 ! Productivity
    integer, parameter :: fnpts=1500 ! Finer grid size
    real*8, dimension (npts) :: ngrid
    real*8, dimension (nepspts) :: epsgrid
    real*8, dimension (fnpts) :: fngrid
    real*8, parameter :: nmin=zero ! Minimum firm size
    real*8, parameter :: nmax=50000.0D+0 ! Maximum firm size
    real*8, dimension (nepspts,nepspts) :: peps ! Transition matrix for productivity
    real*8, dimension (nepspts) :: pinv ! Invariant distribution over productivity

! Cost Function parameters
    real*8, parameter :: z1=3.0D+0 ! Qubic cost
    real*8, parameter :: z2=z1-one ! Model implication

! Calibrated parameters
    real*8, parameter :: mmax=0.99999999999D+0 ! An upper bound on job filling probability
    real*8, parameter :: cv=2.3524*1.0D-4
    real*8, parameter :: cs=10.35415595D+0
    real*8, parameter :: sig=0.21732407D+0
    real*8, parameter :: atech=3.50481081D+0
    real*8, parameter :: b=0.94719736D+0 ! Normalization (from WS model)
! Other model parameters (calibrated)
    real*8, parameter :: alpha=0.677D+0 ! Fixed, RBC literature
    real*8, parameter :: rho=0.590D+0 ! Fixed, Abraham and White (2006)
    real*8, parameter :: beta=one/1.012D+00**(one/12.0D+00) ! Discount factor: quarterly interest rate
    real*8, parameter :: delta=0.015D+00/5.0D+00/4.0D+0 ! Exit prob: 1/6 of separations
    real*8, parameter :: lambda=0.0118D+0/4.0D+0 ! Seperation at zero growth rate (replacement workers)
    real*8, parameter :: psi=(one-delta)*(one-lambda)
    real*8, parameter :: l=log(0.28D+0)/log(1.3D+0/5.3D+0)
    real*8, parameter :: xk=(one/0.28D+0-one)/((one-(one-0.45D+00)**0.25D+0)/(1.3D+0/5.3D+0))**l
    real*8 :: ce, cce
! Policy parameters
    real*8, parameter :: pols=0.00D+0
    real*8, parameter :: polf=0.00D+0
    real*8 :: s, tau
! Endogenous variables
    real*8 :: mu, xmu
! Auxilary varibles
    real*8 :: gdiff, gdiff2, gdiff3, curdlt, curn, mmin, mmint
    integer :: gcount, cureps, dcount
! Approximation error tolerance
    real*8, parameter :: tol=1.0D-8
    real*8, parameter :: mtol=1.0D-8
    real*8, parameter :: dtol=1.0D-8
    real*8, parameter :: gtol=1.0D-8
! Policy functions
    real*8, dimension(npts,nepspts) :: g_dlt, g_env
    real*8, dimension(4,nepspts) :: g_ncr
    integer, dimension(4,nepspts) :: g_ncri, fg_ncri
    real*8, dimension(fnpts,nepspts) :: fg_dlt, fg_env, fg_np, fg_m, &
        fg_v, fg_r, fg_f, fg_dist, nwages, pwages, fg_j, fg_theta
! Cross-section variables
    integer, parameter :: ngrbin=194
    real*8, dimension(ngrbin+1) :: grbin
    real*8, dimension(ngrbin) :: grhrs, grseps, gruvcn, grvy, gremp
    real*8, dimension(fnpts,nepspts) :: dfix, npfix, rfix, ffix, &
              nppol, rpol, fpol
    real*8, dimension(fnpts) :: nfix
! Cross-section variables
    integer, parameter :: ndpts= 1000000
    real*8, dimension(ndpts,5) :: datafw
end module params

program policy_kk
    use params
    implicit none
    integer :: j
    real*8 :: step, zbrent
    interface
      double precision function calfun1(x)
         implicit none
         real*8, intent(in) :: x
      end function calfun1
    end interface


! Create grid points for firm size (ngrid)
ngrid(1)=nmin
ngrid(npts)=nmax
step=(ngrid(npts)-ngrid(1))/dble(real(npts-2))
do j=2,npts
   ngrid(j)=ngrid(j-1)+step
enddo
    g_dlt(:,:)=zero
    g_env(:,:)=zero
    g_ncr(1,:)=one
    g_ncr(2,:)=one+one
    g_ncr(3,:)=one+one+one
    fg_j(:,:)=zero


! Initialize policy functions and the envelope condition
!open(1,file='ngrid.txt',form='formatted')
!open(2,file='gdlt.txt',form='formatted')
!open(3,file='genv.txt',form='formatted')
!open(4,file='gncr.txt',form='formatted')
!open(41,file='fgj.txt',form='formatted')
!do j=1,npts
!   read(1,'(f15.8)') ngrid(j)
!   read(2,'(25f15.8)') g_dlt(j,:)
!   read(3,'(25f15.8)') g_env(j,:)
!enddo
!read(4,'(25f15.8)') g_ncr(1,:)
!read(4,'(25f15.8)') g_ncr(2,:)
!read(4,'(25f15.8)') g_ncr(3,:)
!do j=1, fnpts-1
!   read(41,'(25f15.8)') fg_j(j,:)
!enddo
!close(1)
!close(2)
!close(3)
!close(4)
!close(41)
    fg_dist(:,:)=zero
    fg_dist(1,1)=one
!open(42,file='grbin.txt',form='formatted')
!do j=1,ngrbin+1
!   read(42,'(f10.4)') grbin(j)
!enddo
!close(42)

    call e_grid()
! Step 1- Calculate ce at targeted mu
    ce=zero
    tau=zero
    s=zero
    mu=0.05163679D+0
    step=calfun1(mu)
    ce=cce
    dfix=fg_dist
    nfix=fngrid
    npfix=fg_np
    rfix=fg_r
    ffix=fg_f
    print*, 'ce=', ce
! Set policy value and calculate equilibrium
    s=pols*b
    tau=polf*b
    mu=zbrent(calfun1,0.01D+0,0.1D+0,tol)
    step=calfun1(mu)
    call sim_agg()
    call sim_cross()
    call monthly_simwt()

    open(5,file='peps.txt',form='formatted')
    open(7,file='fngrid.txt',form='formatted')
    open(71,file='ngrid.txt',form='formatted')
    open(8,file='fgdlt.txt',form='formatted')
    open(81,file='gdlt.txt',form='formatted')
    open(9,file='fgdist.txt',form='formatted')
    open(91,file='genv.txt',form='formatted')
    open(92,file='gncr.txt',form='formatted')
    open(10,file='fgnp.txt',form='formatted')
    open(11,file='fgm.txt',form='formatted')
    open(12,file='fgv.txt',form='formatted')
    open(13,file='fgf.txt',form='formatted')
    open(131,file='fgr.txt',form='formatted')
    open(14,file='fgj.txt',form='formatted')
    open(15,file='fgwgsn.txt',form='formatted')
    open(16,file='fgwgsp.txt',form='formatted')
    open(101,file='grvy.txt',form='formatted')
    do j=1,nepspts
       write(5,'(25f15.8)') peps(j,:), sum(peps(j,:))
    enddo
    write(5,'(25f15.8)') pinv(:), sum(pinv(:))
    do j=1,fnpts-1
       write(7,'(f15.8)') fngrid(j)
       write(8,'(25f15.8)') fg_dlt(j,:)
       write(9,'(25f15.8)') fg_dist(j,:)
       write(10,'(25f15.8)') fg_np(j,:)
       write(11,'(25f15.8)') fg_m(j,:)
       write(12,'(25f15.8)') fg_v(j,:)
       write(13,'(25f15.8)') fg_f(j,:)
       write(131,'(25f15.8)') fg_r(j,:)
       write(14,'(25f15.8)') fg_j(j,:)
       write(15,'(25f15.8)') nwages(j,:)
       write(16,'(25f15.8)') pwages(j,:)
    enddo
    write(14,'(25f15.8)') fg_j(fnpts,:)
    do  j=1,npts
       write(71,'(f15.8)') ngrid(j)
       write(81,'(25f15.8)') g_dlt(j,:)
       write(91,'(25f15.8)') g_env(j,:)
    enddo
    write(92,'(25f15.8)') g_ncr(1,:)
    write(92,'(25f15.8)') g_ncr(2,:)
    write(92,'(25f15.8)') g_ncr(3,:)
    do j=1,ngrbin
       write(101,'(6f10.5)') grbin(j), grhrs(j), grseps(j), gruvcn(j), grvy(j), gremp(j)
    enddo
    close(5)
    close(7)
    close(8)
    close(9)
    close(10)
    close(11)
    close(12)
    close(13)
    close(131)
    close(14)
    close(15)
    close(16)
    close(71)
    close(81)
    close(91)
    close(92)
    close(101)


    return

end program policy_kk

subroutine sim_agg()
   use params
   implicit none
   real*8 :: uend, mavg, uvcn, ubar, vbar, hrs, &
             sepf, sepl, sepd, ybar, yubar, nbar, emp, &
             wnbar, wpbar, scost, &
             tpols, tpolf
   real*8 :: n, np, m, v, thtv, r, f, wgt, e
   integer :: j, k


   mavg=zero
   vbar=zero
   uvcn=zero
   ubar=zero
   hrs=zero
   sepf=zero
   sepl=zero
   sepd=zero
   ybar=zero
   yubar=zero
   nbar=zero
   emp=zero
   wnbar=zero
   wpbar=zero
   scost=zero
   tpols=zero
   tpolf=zero
   do j=1,nepspts
      do k=1,fnpts
         if (fg_dist(k,j).eq.zero) cycle
         wgt=fg_dist(k,j)
         e=epsgrid(j)
         n=fngrid(k)
         np=fg_np(k,j)
         m=fg_m(k,j)
         v=fg_v(k,j)
         thtv=fg_theta(k,j)*v
         r=fg_r(k,j)
         f=fg_f(k,j)
! Calculate aggregates
         mavg=mavg+wgt*m*v
         vbar=vbar+wgt*v
         ubar=ubar+wgt*thtv
         uvcn=uvcn+wgt*v*(one-m)
         hrs=hrs+wgt*r
         sepf=sepf+wgt*f
         sepl=sepl+wgt*n*lambda
         sepd=sepd+wgt*n*delta
         ybar=ybar+wgt*e*np**alpha
         nbar=nbar+wgt*n
         emp=emp+wgt*(n+np)/2.0D+0
         wnbar=wnbar+wgt*min((one-lambda)*n,np)*nwages(k,j)
         wpbar=wpbar+wgt*r*pwages(k,j)
         scost=scost+wgt*(cv*v+cs*v**z1/z1/(one+(one-lambda)*n)**z2)
         tpols=tpols+wgt*s*r
         tpolf=tpolf+wgt*tau*f
      enddo
   enddo
   xmu=one/(nbar+ubar)
   uend=one-xmu*nbar
   yubar=b*(one-xmu*((one-delta-lambda)*nbar+hrs))
   open(501,file='agg_stats.txt',form='formatted')
5011 format(f25.10)
   write(501,5011) tau/b
   write(501,5011) mu
   write(501,5011) xmu
   write(501,5011) uend
   write(501,5011) mavg/vbar
   write(501,5011) vbar
   write(501,5011) uvcn
   write(501,5011) nbar
   write(501,5011) emp
   write(501,5011) hrs
   write(501,5011) sepf
   write(501,5011) sepl
   write(501,5011) sepd
   write(501,5011) ybar
   write(501,5011) yubar
   write(501,5011) wnbar
   write(501,5011) wpbar
   write(501,5011) scost
   write(501,5011) tpols
   write(501,5011) tpolf
   close(501)
   return
end subroutine sim_agg

subroutine sim_cross()
   use params
   implicit none
   integer :: k, j, jlo
   real*8 :: wgt, gr, np, n, w
   integer, parameter :: nbin=ngrbin
   real*8, dimension (nbin) :: dss, dpol
   real*8, dimension (nbin+1) :: pgrbin

   pgrbin=grbin
   dss(:)=zero
   dpol(:)=zero

   do k=1, fnpts
      jlo=1
      call hunt(fngrid,fnpts,nfix(k),jlo,wgt)
      do j=1, nepspts
          nppol(k,j)=wgt*fg_np(jlo,j)+(one-wgt)*fg_np(jlo+1,j)
          rpol(k,j)=wgt*fg_r(jlo,j)+(one-wgt)*fg_r(jlo+1,j)
          fpol(k,j)=wgt*fg_f(jlo,j)+(one-wgt)*fg_f(jlo+1,j)
      enddo
   enddo

   ! Find initial growth rate
   do j=1, nepspts
      do k=1, fnpts
         w=dfix(k,j)
         if (w.eq.zero) cycle
         n=nfix(k)
         np=npfix(k,j)
         gr=2.0D+0*(np-n)/(np+n)
         jlo=1
         call hunt(pgrbin,nbin+1,gr,jlo,wgt)
         dss(jlo)=dss(jlo)+w
      enddo
   enddo

   ! Find after policy growth rate
   do j=1, nepspts
      do k=1, fnpts
         w=dfix(k,j)
         if (w.eq.zero) cycle
         n=nfix(k)
         np=nppol(k,j)
         gr=2.0D+0*(np-n)/(np+n)
         jlo=1
         call hunt(pgrbin,nbin+1,gr,jlo,wgt)
         dpol(jlo)=dpol(jlo)+w
      enddo
   enddo
   ! Write cumulative distribution function
   open(502,file='cross_stats.txt',form='formatted')
5021 format (f8.4, 5f15.10)

   write(502,5021) pgrbin(1), dss(1), dpol(1), dss(1), dpol(1)
   w=dss(1)
   n=dpol(1)
   do j=2, nbin
      w=w+dss(j)
      n=n+dpol(j)
      write(502,5021) pgrbin(j), dss(j), dpol(j), w, n
   enddo
   close(502)

! Calculate wt changes after policy

   open(503,file='cross_data.txt',form='formatted')
   do j=1, nepspts
      do k=1, fnpts
      if (dfix(k,j).lt.1.0D-7) cycle
      write(503,'(6f20.7)') dfix(k,j), nfix(k), npfix(k,j), &
              rfix(k,j), nppol(k,j), rpol(k,j)
      enddo
   enddo
   close(503)

   return

end subroutine sim_cross

subroutine monthly_simwt()
    use params
    implicit none
    integer, parameter :: period=4
    real*8, dimension(ngrbin) :: grn, grn2, grd, grd2
    integer, parameter :: nwtbin=6
    real*8, parameter :: wt1=0.619D+0!0.244D+0
    real*8, dimension (nwtbin) :: wtbin
    real*8, dimension(nwtbin) :: wthrs, wthrs2, wtseps, wtseps2, wtemp, wtemp2
    real*8, dimension(period) :: hrs, seps
    real*8, dimension(fnpts,nepspts) :: npfunc, rfunc, ffunc
    integer :: j, k, jlo
    real*8 :: n0, hrm, hrm2, sepm, sepm2, np, np2,  &
       w, emp, emp2, empt, empb, wgt, gr
    real*8, dimension(:,:), allocatable :: dataw
    real*8, dimension(:), allocatable :: datawt
    integer, dimension (:), allocatable :: op


    print*, 'simulating...'
    dcount=0
    npfunc=npfix
    rfunc=rfix
    ffunc=ffix
    do j=1,nepspts
    do k=1,fnpts
       n0=nfix(k)
       w=dfix(k,j)
       call sim_fw(1 ,period, j, n0, w, hrs, seps, n0, &
            npfunc, rfunc, ffunc)
    enddo
    enddo
    print*, 'Simulated first part'
    allocate(dataw(dcount,8))
    dataw(:,1:5)=datafw(1:dcount,:)
    ! do calculations after policy
    print*, 'Simulating second part'
    datafw(:,:)=zero
    dcount=0
    npfunc=nppol
    rfunc=rpol
    ffunc=fpol
    do j=1,nepspts
    do k=1,fnpts
       n0=nfix(k)
       w=dfix(k,j)
       call sim_fw(1 ,period, j, n0, w, hrs, seps, n0, &
                   npfunc, rfunc, ffunc)
    enddo
    enddo
    print*, 'Simulated second part'
    dataw(:,6:8)=datafw(1:dcount,3:5)
    print*, 'work data created...'
    empt=zero
    grn(:)=zero
    grd(:)=zero
    grn2(:)=zero
    grd2(:)=zero
    do j=1, dcount
        w=dataw(j,1)
        n0=dataw(j,2)
        np=dataw(j,5)
        np2=dataw(j,8)
        empt=empt+w*(n0+np)/2.0D+0
        gr=2.0D+0*(np-n0)/(np+n0)
        jlo=1
        call hunt(grbin,ngrbin+1,gr,jlo,wgt)
        grn(jlo)=grn(jlo)+w*(np-n0)
        grd(jlo)=grd(jlo)+w*(np+n0)/2.0D+0
        grn2(jlo)=grn2(jlo)+w*(np2-n0)
        grd2(jlo)=grd2(jlo)+w*(np2+n0)/2.0D+0
    enddo
    open(102,file='grpol.txt',form='formatted')
    do j=1,ngrbin
       write(102,'(4f15.10)') grbin(j), grn(j)/grd(j), &
       grn2(j)/grd2(j), grn2(j)/grd2(j)-grn(j)/grd(j)
    enddo
    close(102)
    ! create worker turnover bins
    wtbin(1)=wt1
    w=(one-wtbin(1))/dble(real(nwtbin-1))
    do j=2,nwtbin
       wtbin(j)=wtbin(j-1)+w
    enddo
    wtbin(nwtbin)=one
    allocate(datawt(dcount))
    allocate(op(dcount))
    empt=zero
    do j=1, dcount
        hrm=dataw(j,3)
        sepm=dataw(j,4)
        n0=dataw(j,2)
        np=dataw(j,5)
        datawt(j)=2.0D+0*(hrm+sepm)/(n0+np)
        empt=empt+dataw(j,1)*(n0+np)/2.0D+0
    enddo
    op=(/(j,j=1,dcount)/)
    call ssort(datawt,op,dcount)
    wthrs(:)=zero
    wtseps(:)=zero
    wtemp(:)=zero
    wthrs2(:)=zero
    wtseps2(:)=zero
    wtemp2(:)=zero
    empb=zero
    jlo=1
    do j=1, dcount
        w=dataw(op(j),1)
        hrm=dataw(op(j),3)
        sepm=dataw(op(j),4)
        hrm2=dataw(op(j),6)
        sepm2=dataw(op(j),7)
        n0=dataw(op(j),2)
        np=dataw(op(j),5)
        np2=dataw(op(j),8)
        emp=(n0+np)/2.0D+0
        emp2=(n0+np2)/2.0D+0
        if ((empb+wtemp(jlo)+w*emp)/empt.gt.wtbin(jlo)) then
           empb=empb+wtemp(jlo)
           jlo=jlo+1
        endif
        wthrs(jlo)=wthrs(jlo)+w*hrm
        wthrs2(jlo)=wthrs2(jlo)+w*hrm2
        wtseps(jlo)=wtseps(jlo)+w*sepm
        wtseps2(jlo)=wtseps2(jlo)+w*sepm2
        wtemp(jlo)=wtemp(jlo)+w*emp
        wtemp2(jlo)=wtemp2(jlo)+w*emp2
    enddo
    do j=1, nwtbin
       wthrs(j)=wthrs(j)/wtemp(j)
       wthrs2(j)=wthrs2(j)/wtemp2(j)
       wtseps(j)=wtseps(j)/wtemp(j)
       wtseps2(j)=wtseps2(j)/wtemp2(j)
       wtemp(j)=wtemp(j)/empt
    enddo
    open(103,file='wtpol.txt',form='formatted')
    do j=1,nwtbin
       write(103,'(6f10.5)') wtbin(j), wthrs(j), wtseps(j), wthrs2(j), wtseps2(j), wtemp(j)
    enddo
    close(103)
    deallocate(datawt)
    deallocate(op)
    print*, 'wt after policy: written on wtpol.txt'
    deallocate(dataw)
    print*, 'simulation ended'
    return
end


recursive subroutine sim_fw(t,period, ceps, cn, &
                     dx, hrs, seps, n0, npfunc, rfunc, ffunc)
   use params
   implicit none
   integer, intent(in) :: t, period, ceps
   real*8, intent(in) :: cn, dx
   real*8, intent(inout) :: n0
   real*8, dimension(fnpts, nepspts) :: npfunc, rfunc, ffunc
   integer :: jlo, k
   real*8, dimension (period), intent(inout) :: hrs, seps
   real*8 :: wgt, dxx, np

   if (dx.le.1.0D-7) return
   jlo=1
   call hunt(nfix,fnpts,cn,jlo,wgt)
   if(t.eq.1) n0=cn
   if(t.eq.period) then ! Calculate cross section vacancy yield
      ! Final period policy choice and productivity
      np=wgt*npfunc(jlo,ceps)+(one-wgt)*npfunc(jlo+1,ceps)
      hrs(t)=wgt*rfunc(jlo,ceps)+(one-wgt)*rfunc(jlo+1,ceps)
      seps(t)=lambda*cn+wgt*ffunc(jlo,ceps) &
             +(one-wgt)*ffunc(jlo+1,ceps)
      dcount=dcount+1
      datafw(dcount,:)=(/dx, n0, sum(hrs(1:period)), sum(seps(1:period)), np/)
      return
   else
          ! Calculate the current policy values for (n,eps)
          np=wgt*npfunc(jlo,ceps)+(one-wgt)*npfunc(jlo+1,ceps)
          hrs(t)=wgt*rfunc(jlo,ceps)+(one-wgt)*rfunc(jlo+1,ceps)
          seps(t)=lambda*cn+wgt*ffunc(jlo,ceps) &
                  +(one-wgt)*ffunc(jlo+1,ceps)
          do k=1,nepspts
             dxx=dx*peps(ceps,k)
             call sim_fw(t+1,period, k, np, &
                  dxx, hrs, seps, n0, npfunc, rfunc, ffunc)
          enddo
   endif
end


double precision function calfun1(x)
    use params
    implicit none
    real*8, intent(in) :: x
    integer :: j, time(8)

    call date_and_time(values=time)
! Step 1: Initialize eqm value of mu
    mu=x

! Step 2: Solve firms' problem and calculate staionary distribution
    call firm_opt()
    call calcdist(fg_dist)
! Step 3: Calculate firm value at entry
    call firm_val()
    cce=zero
    do j=1, nepspts
       cce=cce+pinv(j)*fg_j(1,j)
    enddo
! Step 5: Calculate residual
   calfun1=cce-ce
   ! Write output
   gcount=gcount+1
   print '(a43)', '+++++++++++++++++++++++++++++++++++++++++++'
   print '(a10,i5,5x,a9, i2,a1,i2,a1,i2)', 'Iteration=', gcount
   print '(a11, i2,a1,i2,a1,i2)',  'Start Time:', time(5), ':', time(6), ':', time(7)
   call date_and_time(values=time)
   print '(a11, i2,a1,i2,a1,i2)', 'End Time:  ', time(5), ':', time(6), ':', time(7)
   print*, ' '
   print '(a4,f15.8)','SSE=', calfun1
   print '(a15,f15.8)','Firm_Opt error=', gdiff
   print '(a15,f15.8)','Dist sum=      ', gdiff2
   print '(a15,f15.8)','Firm_Val error=', gdiff3
   print*, ' '
   print '(a19)', 'Current eqm value:'
   write(6,12) ' mu=', mu
   print*, ' '
   print '(a58)', 'Moment       Current        Target         Diff           '
   print '(a58)', '-------------------------------------------'
   write(6,13) 'Firm Entry  :', cce, ce, cce-ce
   print '(a43)', '+++++++++++++++++++++++++++++++++++++++++++'
   print*, ' '
12 format(a4,f15.8)
13 format(a13,3f15.8)

   return

end function calfun1

subroutine e_grid()
! Create the productivity grid, calculate the transition
! matrix and the unconditional distribution of productivities.
    use params
    implicit none
    real*8, dimension(nepspts,nepspts) :: pit
    integer :: j

   call tauchen (nepspts,epsgrid,peps,rho,sig)
   epsgrid =atech*dexp(epsgrid)
   peps=peps/52.0D+0
   do j=1,nepspts
      peps(j,j)=51.0D+0/52.0D+0+peps(j,j)
   enddo

   ! Correct for probability sum equals to 1
   do j=1,nepspts
      peps(j,j)=one-sum(peps(j,1:j-1))-sum(peps(j,j+1:nepspts))
   enddo
   pit = peps
   do j = 1,100000
      pit = matmul(pit,peps)
   end do
   pinv = pit(1,:)
   ! Correct for probability sum equals to 1
   pinv(nepspts/2)=one-sum(pinv(1:nepspts/2-1))-sum(pinv(nepspts/2+1:nepspts))

    return
end


subroutine firm_opt()
   use params
   implicit none
   integer, parameter :: kmax=500
   integer :: j, k, kkk, jlo, t
   real*8, dimension (npts) :: ngrid0
   integer, dimension (npts) :: op
   integer, dimension (fnpts) :: fop
   real*8 :: step, zbrent, wgt, m, envfunc, diff, ediff, &
             ndiff, ddiff
   real*8, dimension(npts,nepspts) :: g_dlt0, g_env0
   real*8, dimension(4,nepspts) :: g_ncr0
    interface
       double precision function root1(x)
         implicit none
         real*8, intent(in) :: x
       end function root1
    end interface
    interface
       double precision function root2(x)
         implicit none
         real*8, intent(in) :: x
       end function root2
    end interface
    interface
       double precision function root3(x)
         implicit none
         real*8, intent(in) :: x
       end function root3
    end interface
    interface
       double precision function rootd(x)
         implicit none
         real*8, intent(in) :: x
       end function rootd
    end interface
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface

curdlt=zero
curn=one
mmint=zbrent(rootm,zero,mmax,mtol)
mmin=mmint/2.0D+0
! Guess policy functions
g_env0=g_env
g_dlt0=g_dlt
g_ncr0=g_ncr
ngrid0=ngrid
! Iterate policy functions
do kkk=1,kmax
! Step 1: Update policy functions with the new solution
g_env=g_env0
g_dlt=g_dlt0
g_ncr=g_ncr0
ngrid=ngrid0
! Step 2: Solve for critical n values
!     i- Find the largest size firm with gdlt=0,
!        given epsilon
do j=1,nepspts
cureps=j
g_ncr(3,j)=zbrent(root3,nmin,nmax,gtol)
g_ncr(4,j)=(one-lambda)*g_ncr(3,j)
enddo
!    ii- Find the smallest size firm with gdlt=0,
!        given epsilon
do j=1,nepspts
cureps=j
g_ncr(2,j)=zbrent(root2,nmin,g_ncr(3,j)+1.0D+0,gtol)
enddo
!   iii- Find the firm size with gdlt=lambda*n,
!        given epsilon
do j=1,nepspts
cureps=j
g_ncr(1,j)=zbrent(root1,nmin,g_ncr(2,j)+1.0D+0,gtol)
enddo
! Step 4: Create a new grid (ngrid0) for firm size
!     This grid now includes all the critical values
!     Also calculate the location of these points
!     to be used in profit maximization
ngrid0(1)=nmin
ngrid0(npts)=nmax
ngrid0(npts-4*nepspts-1)=g_ncr(3,nepspts)+1.0D-0
step=2.0D+0*(ngrid0(npts-4*nepspts-1)-nmin)/dble(real(npts-4*nepspts-2))/dble(real(npts-4*nepspts-1))
do j=2,npts-4*nepspts-1
   ngrid0(j)=ngrid0(j-1)+dble(real(j-1))*step
enddo
t=npts-4*nepspts
   do j=1,4
      do k=1,nepspts
         ngrid0(t)=g_ncr(j,k)
         t=t+1
      enddo
   enddo
op=(/(j,j=1,npts)/)
call ssort(ngrid0,op,npts)
do j=1,4
   do k=1,nepspts
      jlo=1
      call hunt(ngrid0,npts,g_ncr(j,k),jlo,wgt)
      if (wgt.lt.tol) then
         g_ncri(j,k)=jlo+1
      else
         g_ncri(j,k)=jlo
      endif
   enddo
enddo
! Step 6: Solve profit maximization for hiring firms
!         and solve for the envelope condition
do j=1,nepspts
cureps=j
! Hiring firms
do k=1,g_ncri(2,j)-1
curn=ngrid0(k)
g_dlt0(k,j)=zbrent(rootd,zero,g_ncr(2,j)+1.0D+0,gtol)
curdlt=g_dlt0(k,j)
m=zbrent(rootm,mmin,mmax,mtol)
g_env0(k,j)=envfunc(m,ngrid0(k),g_dlt0(k,j),j)
enddo
! Adjustment for first cut-off
curn=ngrid0(g_ncri(1,j))
g_dlt0(g_ncri(1,j),j)=lambda*curn
curdlt=g_dlt0(g_ncri(1,j),j)
m=zbrent(rootm,mmin,mmax,mtol)
g_env0(g_ncri(1,j),j)=envfunc(m,curn,g_dlt0(g_ncri(1,j),j),j)
! Second cut-off to third cut-off (including 4th critical point)
do k=g_ncri(2,j),g_ncri(3,j)-1
curn=ngrid0(k)
g_dlt0(k,j)=zero
m=mmint
g_env0(k,j)=envfunc(m,curn,g_dlt0(k,j),j)
enddo
! Adjustment for third cut-off
g_dlt0(g_ncri(3,j),j)=zero
g_env0(g_ncri(3,j),j)=envfunc(m,ngrid0(k),-one,j)
! Third cut-off onward
do k=g_ncri(3,j)+1,npts
g_dlt0(k,j)=(one-lambda)*(g_ncr(3,j)-ngrid0(k))
g_env0(k,j)=envfunc(m,ngrid0(k),g_dlt0(k,j),j)
enddo
enddo
! Check for convergence
ndiff=maxval(abs(ngrid-ngrid0))
ddiff=maxval(abs(g_dlt-g_dlt0))
ediff=maxval(abs(g_env-g_env0))
diff=max(ndiff,ddiff,ediff)
if (diff.le.tol) exit
! End of iteration
enddo
gdiff=diff
g_dlt=g_dlt0
g_env=g_env0
ngrid=ngrid0
! Step 7: Create a finer grid
fngrid(1)=nmin
fngrid(fnpts)=nmax
fngrid(fnpts-4*nepspts-1)=g_ncr(4,nepspts)+1.0D-0
step=2.0D+0*(fngrid(fnpts-4*nepspts-1)-nmin)/dble(real(fnpts-4*nepspts-2))/dble(real(fnpts-4*nepspts-1))
do j=2,fnpts-4*nepspts-1
   fngrid(j)=fngrid(j-1)+dble(real(j-1))*step
enddo
t=fnpts-4*nepspts
   do j=1,4
      do k=1,nepspts
         fngrid(t)=g_ncr(j,k)
         t=t+1
      enddo
   enddo
fop=(/(j,j=1,fnpts)/)
call ssort(fngrid,fop,fnpts)
do j=1,4
   do k=1,nepspts
      jlo=1
      call hunt(fngrid,fnpts,g_ncr(j,k),jlo,wgt)
      if (wgt.lt.tol) then
         fg_ncri(j,k)=jlo+1
      else
         fg_ncri(j,k)=jlo
      endif
   enddo
enddo

! Step 8: Carry the policy functions over the finer grid

do j=1,nepspts
! Write in hiring firms decision
t=g_ncri(2,j)
do k=1,fg_ncri(2,j)-1
jlo=1
call hunt(ngrid(1:t),t,fngrid(k),jlo,wgt)
fg_dlt(k,j)=wgt*g_dlt(jlo,j)+(one-wgt)*g_dlt(jlo+1,j)
fg_env(k,j)=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
enddo
! correct no growth firms
fg_dlt(fg_ncri(1,j),j)=lambda*fngrid(fg_ncri(1,j))
fg_env(fg_ncri(1,j),j)=g_env(g_ncri(1,j),j)
! Shrinking firms
t=g_ncri(3,j)
do k=fg_ncri(2,j),fg_ncri(3,j)-1
jlo=1
call hunt(ngrid(1:t),t,fngrid(k),jlo,wgt)
fg_dlt(k,j)=zero
fg_env(k,j)=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
enddo
! Adjustment for the third cut-off
fg_dlt(fg_ncri(3,j),j)=zero
fg_env(fg_ncri(3,j),j)=envfunc(m,fngrid(k),-one,j)
do k=fg_ncri(3,j)+1,fnpts
fg_dlt(k,j)=(one-lambda)*(fngrid(fg_ncri(3,j))-fngrid(k))
fg_env(k,j)=envfunc(m,fngrid(k),fg_dlt(k,j),j)
enddo
enddo
! Write in other policy functions and workers value function
do j=1,nepspts
cureps=j
do k=1,fnpts
curn=fngrid(k)
curdlt=fg_dlt(k,j)
fg_np(k,j)=(one-lambda)*curn+fg_dlt(k,j)
if(k.ge.fg_ncri(2,j).and.k.le.fg_ncri(3,j)) fg_np(k,j)=(one-lambda)*curn
if (fg_dlt(k,j).gt.zero) then
fg_m(k,j)=zbrent(rootm,mmin,mmax,mtol)
fg_theta(k,j)=fg_m(k,j)*(xk/(one-fg_m(k,j)**l))**(one/l)
nwages(k,j)=b+mu
pwages(k,j)=fg_theta(k,j)*mu/fg_m(k,j)+b
fg_v(k,j)=fg_dlt(k,j)/fg_m(k,j)
fg_r(k,j)=fg_dlt(k,j)
fg_f(k,j)=zero
else
fg_m(k,j)=mmint
fg_theta(k,j)=zero
nwages(k,j)=b+mu
pwages(k,j)=b+mu
fg_v(k,j)=zero
fg_r(k,j)=zero
fg_f(k,j)=-fg_dlt(k,j)
endif

enddo
enddo

return
end subroutine firm_opt



subroutine calcdist(dist1)
    use params
    implicit none
    real*8, dimension(fnpts,nepspts), intent(inout) :: dist1
    real*8, dimension(fnpts,nepspts) :: dist0
    integer :: t, j, k, jj, jlo
    integer, parameter :: tmax=3000
    real*8 :: wgt, diff

    do t=1,tmax
    dist0=dist1
    dist1(:,:)=zero
       do j=1,nepspts
          dist1(1,j)=delta*pinv(j)
          do k=1,fnpts
             if (dist0(k,j).eq.zero) cycle
             jlo=1
             call hunt(fngrid,fnpts,fg_np(k,j),jlo,wgt)
             do jj=1,nepspts
                 dist1(jlo,jj)=dist1(jlo,jj)+peps(j,jj)*(one-delta)*dist0(k,j)*wgt
                 dist1(jlo+1,jj)=dist1(jlo+1,jj)+peps(j,jj)*(one-delta)*dist0(k,j)*(one-wgt)
             enddo
          enddo
       enddo
    diff=maxval(abs(dist1-dist0))
    if (diff.le.1.0D-10) exit
    enddo
    gdiff2=sum(dist1)
    return

end subroutine calcdist

subroutine firm_val()
   use params
   implicit none
   integer, parameter :: tmax=20000
   integer :: k, j, jj, t, jlo
   real*8, dimension (fnpts,nepspts) :: fg_j0
   real*8 :: n, np, m, v, f, e, r, ej, wn, wp, &
             wgt, diff

   fg_j0=fg_j
   do t=1,tmax
   fg_j=fg_j0
      do j=1, nepspts
         do k=1, fnpts
            e=epsgrid(j)
            n=fngrid(k)
            np=fg_np(k,j)
            m=fg_m(k,j)
            v=fg_v(k,j)
            r=fg_r(k,j)
            f=fg_f(k,j)
            wn=nwages(k,j)
            wp=pwages(k,j)
            ej=zero
            jlo=1
            call hunt(fngrid,fnpts,np,jlo,wgt)
            do jj=1,nepspts
               ej=ej+beta*(one-delta)*peps(j,jj) &
                  *(wgt*fg_j(jlo,jj)+(one-wgt)*fg_j(jlo+1,jj))
            enddo
            fg_j0(k,j)= -cv*v-cs*v**z1/z1/(one+(one-lambda)*n)**z2 &
               +e*np**alpha-wn*(one-lambda)*n-wp*r+ej+r*s-tau*f
         enddo
      enddo
      diff=maxval(abs(fg_j-fg_j0))
      if (diff.le.tol) exit
   enddo
   gdiff3=diff
   return
end subroutine firm_val

double precision function root1(x)
    use params
    implicit none
    real*8, intent(in) :: x
    real*8 :: e, m, ej, wgt, denv
    real*8 :: zbrent
    integer :: j, jlo
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface
    e=epsgrid(cureps)
    curdlt=lambda*x
    curn=x
    m=zbrent(rootm,mmin,mmax,mtol)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,x,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    root1=alpha*e*x**(alpha-one)+beta*(one-delta)*ej &
          -b+s-mu*(xk/(one-m**l))**(one/l)/(one-m**l)
    return

end function root1

double precision function root2(x)
    use params
    implicit none
    real*8, intent(in) :: x
    real*8 :: e, m, ej, wgt, denv
    real*8 :: zbrent
    integer :: j, jlo
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface

    e=epsgrid(cureps)
    curdlt=zero
    curn=x
    m=zbrent(rootm,mmint,mmax,mtol)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,(one-lambda)*x,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    root2=alpha*e*((one-lambda)*x)**(alpha-one) &
          -b+s-mu*(xk/(one-m**l))**(one/l)/(one-m**l) &
          +beta*(one-delta)*ej
    return

end function root2


double precision function root3(x)
    use params
    implicit none
    real*8, intent(in) :: x
    real*8 :: e, ej, wgt, denv
    integer :: j, jlo

    e=epsgrid(cureps)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,(one-lambda)*x,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    root3=alpha*e*((one-lambda)*x)**(alpha-one) &
          -b-mu+tau+beta*(one-delta)*ej
    return

end function root3

double precision function rootd(dx)
    use params
    implicit none
    real*8, intent(in) :: dx
    real*8 :: e, np, m, ej, wgt, denv
    real*8 :: zbrent
    integer :: j, jlo
    interface
       double precision function rootm(x)
         implicit none
         real*8, intent(in) :: x
       end function rootm
    end interface

    e=epsgrid(cureps)
    curdlt=dx
    np=(one-lambda)*curn+dx
    m=zbrent(rootm,mmin,mmax,mtol)
    ej=zero
    jlo=1
    call hunt(ngrid,npts,np,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo

    rootd=alpha*e*np**(alpha-one)+beta*(one-delta)*ej &
          -b+s-mu*(xk/(one-m**l))**(one/l)/(one-m**l)

    return

end function rootd


double precision function envfunc(m,x,dx,eps)
! Envelope(n, Dn,eps)
    use params
    implicit none
    integer, intent(in) :: eps
    real*8, intent(in) :: m,x,dx
    real*8 :: ej, wgt, denv, np
    integer :: j, jlo

    np=(one-lambda)*x+dx

    if (dx.lt.zero) then
        envfunc=-tau
    else
    ej=zero
    jlo=1
    call hunt(ngrid,npts,np,jlo,wgt)
    do j=1,nepspts
        denv=wgt*g_env(jlo,j)+(one-wgt)*g_env(jlo+1,j)
        ej=ej+peps(cureps,j)*denv
    enddo
    envfunc=alpha*epsgrid(eps)*np**(alpha-one)+beta*(one-delta)*ej &
            -b-mu+z2*cs*(dx/m)**z1*(one+(one-lambda)*x)**(-z2-one)/z1
    endif
    envfunc=(one-lambda)*envfunc
    return
end function envfunc

double precision function rootm(x)
   use params
   implicit none
   real*8, intent(in) :: x

   rootm=-mu*xk**(one/l)*x**(one+l)/(one-x**l)**((one+l)/l) &
         +cs*(curdlt/x)**(z1-one)*(one+(one-lambda)*curn)**(-z2)+cv

   return
end function rootm



      double precision function zbrent(func,x1,x2,tol)
      implicit none
      real*8 :: x1,x2,tol,zip,fa,fb,a,b,c,fc,d,e,tol1,xm,s,p,q,r
      integer :: iter
      real*8, parameter :: eps = 3.0D-08
      integer, parameter :: itmax = 200
      interface
         double precision function func(x)
         implicit none
         real*8, intent(in) :: x
         end function func
      end interface

      a = x1
      b = x2
      fa = func(a)
      fb = func(b)
      zip = 0.0D+00

      if ((fa .gt. zip .and. fb .gt. zip) .or. (fa .lt. zip .and. fb .lt. zip)) then
         print*, fa, fb, x1, x2
         stop 'root must be bracated'
      endif

      c = b
      fc = fb
      do iter = 1,itmax
         if ((fb .gt. zip .and. fc .gt. zip) .or. (fb .lt. zip .and. fc .lt. zip)) then
            c = a
            fc = fa
            d = b-a
            e = d
         endif
         if (dabs(fc) .lt. dabs(fb)) then
            a = b
            b = c
            c = a
            fa = fb
            fb = fc
            fc = fa
         endif
         tol1 = 2.0D+00*eps*dabs(b)+0.5D+00*tol
         xm = 0.5D+00*(c-b)
         if (dabs(xm) .le. tol1 .or. fb .eq. zip) then
            zbrent = b
            return
         endif
         if (dabs(e) .ge. tol1 .and. dabs(fa) .gt. dabs(fb)) then
            s = fb/fa
            if (a .eq. c) then
               p = 2.0D+00*xm*s
               q = 1.0D+00-s
            else
               q = fa/fc
               r = fb/fc
               p = s*(2.0D+00*xm*q*(q-r)-(b-a)*(r-1.0D+00))
               q = (q-1.0D+00)*(r-1.0D+00)*(s-1.0D+00)
            endif
            if (p .gt. zip) q = -q
            p = dabs(p)
            if (2.0D+00*p .lt. dmin1(3.0D+00*xm*q-dabs(tol1*q),dabs(e*q))) then
               e = d
               d = p/q
            else
               d = xm
               e = d
            endif
         else
            d = xm
            e = d
         endif
         a = b
         fa = fb
         if (dabs(d) .gt. tol1) then
            b = b+d
         else
            b = b+sign(tol1,xm)
         endif
         fb = func(b)
      end do

      zbrent = b
      stop 'zbrent exceeding maximum iterations'
      return
      end


      subroutine hunt(xx,n,x,jlo,wgt)
      implicit none
      integer, intent(in) :: n
      real*8, intent(in), dimension(n) :: xx
      real*8, intent(in) :: x
      integer, intent(out) :: jlo
      real*8, intent(out) :: wgt
      integer :: jhi,inc,jm
      logical :: ascnd

      ascnd = xx(n) .ge. xx(1)
      if (jlo .le. 0 .or. jlo .gt. n) then
         jlo = 0
         jhi = n-1
         goto 3
      endif

      inc = 1

      if (x .ge. xx(jlo) .eqv. ascnd) then
 1       jhi = jlo+inc
         if (jhi .gt. n) then
            jhi = n+1
         elseif (x .ge. xx(jhi) .eqv. ascnd) then
            jlo = jhi
            inc = inc+inc
            goto 1
         endif
      else
         jhi = jlo
 2       jlo = jhi-inc
         if (jlo .lt. 1) then
            jlo = 0
         elseif (x .lt. xx(jlo) .eqv. ascnd) then
            jhi = jlo
            inc = inc+inc
            goto 2
         endif
      endif
 3    if (jhi-jlo .eq. 1) then
         if (x .eq. xx(n)) jlo = n-1
         if (x .eq. xx(1)) jlo = 1
         go to 4
      endif
      jm = (jhi+jlo)/2
      if (x .ge. xx(jm) .eqv. ascnd) then
         jlo = jm
      else
         jhi = jm
      endif
      goto 3

4 continue
      wgt=(xx(jlo+1)-x)/(xx(jlo+1)-xx(jlo))
      return
      end

subroutine tauchen(n,z,p,rho,sig)
    implicit none
    integer :: j, k
    real*8 :: step
    integer, intent(in) :: n
    real*8, intent(in) :: rho, sig
    real*8, dimension(n), intent(out) :: z
    real*8, dimension(n,n), intent(out) :: p
    real*8 :: mu, cm, zmax, zmin, alnorm
    logical :: up
! Construct grid points for z
    zmin = -3.0D+0*sig/dsqrt(1.0D+0-rho**2)
    zmax =  3.0D+0*sig/dsqrt(1.0D+0-rho**2)
    step = (zmax-zmin)/dble(real((n-1)))
    z(1) = zmin
    do j = 2, n
      z(j) = z(j-1)+step
    end do
    z(n) = zmax
! Construct probabilities
    up=.false.
    mu=0.0D+0
    do j = 1, n
        cm = mu + rho*z(j)
        p(j,1) = alnorm((z(1) - cm + 0.5D+0*step)/sig,up)
        do k = 2, n-1
            p(j,k) = alnorm((z(k) - cm + 0.5D+0*step)/sig,up)  &
                     - alnorm((z(k) - cm - 0.5D+0*step)/sig,up)
        end do
        p(j,n) = 1.0D+0 - alnorm((z(n) - cm - 0.5D+0*step)/sig,up)
    end do

    return

end subroutine tauchen

FUNCTION alnorm( x, upper ) RESULT( fn_val )
!  Algorithm AS66 Applied Statistics (1973) vol.22, no.3

!  Evaluates the tail area of the standardised normal curve
!  from x to infinity if upper is .true. or
!  from minus infinity to x if upper is .false.

! ELF90-compatible version by Alan Miller
! Latest revision - 29 November 2001
   IMPLICIT NONE
   INTEGER, PARAMETER     :: dp = SELECTED_REAL_KIND(15, 100)
   REAL(DP), INTENT(IN)   ::  x
   LOGICAL,   INTENT(IN)  ::  upper
   REAL(DP)               ::  fn_val

   !  Local variables
   REAL(DP), PARAMETER   ::  zero=0.0_DP, one=1.0_DP, half=0.5_DP, con=1.28_DP
   REAL(DP)              ::  z, y
   LOGICAL               ::  up

   !  Machine dependent constants
   REAL(DP), PARAMETER  ::  ltone = 7.0_DP, utzero = 18.66_DP
   REAL(DP), PARAMETER  ::  p = 0.398942280444_DP, q = 0.39990348504_DP,   &
                            r = 0.398942280385_DP, a1 = 5.75885480458_DP,  &
                            a2 = 2.62433121679_DP, a3 = 5.92885724438_DP,  &
                            b1 = -29.8213557807_DP, b2 = 48.6959930692_DP, &
                            c1 = -3.8052E-8_DP, c2 = 3.98064794E-4_DP,     &
                            c3 = -0.151679116635_DP, c4 = 4.8385912808_DP, &
                            c5 = 0.742380924027_DP, c6 = 3.99019417011_DP, &
                            d1 = 1.00000615302_DP, d2 = 1.98615381364_DP,  &
                            d3 = 5.29330324926_DP, d4 = -15.1508972451_DP, &
                            d5 = 30.789933034_DP

   up = upper
   z = x
   IF( z < zero ) THEN
      up = .NOT. up
      z = -z
   END IF
   IF( z <= ltone  .OR.  (up  .AND.  z <= utzero) ) THEN
      y = half*z*z
      IF( z > con ) THEN
         fn_val = r*EXP( -y )/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
      ELSE
         fn_val = half - z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
      END IF
   ELSE
      fn_val = zero
   END IF

   IF( .NOT. up ) fn_val = one - fn_val
   RETURN
END FUNCTION alnorm
